/*
    control++ : System configuration tool
    Copyright (C) 2017-2019 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef confH
#define confH

#include "libcontrol++.h"
#include "text.h"
#include "filesystem.h"

#define CONF_MEM_STEP  4    // Step of memory allocation for conf DOM

//-----------------------------------------------------------------------------

// Check if assgn has no values
bool EmptyAssgn(const vector<string> & assgn);

//-----------------------------------------------------------------------------

// Section of configuration file
class TConfSec{
    private:
        string name;   // Section name
        // Assignments of the section
        // - 1st index - assignment number
        // - 0th element of the 2nd index - variable name of the current assgn
        // - {1, 2, ..} elements of the 2nd index - values of the current assgn
        vector<vector<string> > assgn;
    public:
        // Data access
        const string& Name(void) const{return name;}
        const vector<vector<string> >& Assgn(void) const{return assgn;}
        vector<string> VarAssgn(const string &var_name) const;
        string VarSingleAssgn(const string &var_name) const;
        // Modifiers
        void WriteName(const string &new_name=""){
            name = new_name;
            }
        void WriteAssgn(const vector<vector<string> > &new_assgn){
            assgn = new_assgn;
            }
        void AddAssgn(const vector<string> &new_assgn){
            assgn.push_back(new_assgn);
            }
        bool ChangeAssgn(const vector<string> &new_assgn);
        bool ChangeAssgn(const vector<string> &new_assgn, int i);

        void Clear(void){
            name.clear();
            assgn.clear();
            }
        TConfSec(void){}
        TConfSec(const string &sec_name){
            name = sec_name;
            }
        ~TConfSec(void){}
    };

//-----------------------------------------------------------------------------

// Configuration file DOM
class TConfDOM{
    private:
        vector<TConfSec> sec;   // Sectors of the configuration file
    public:
        // Data access
        const vector<TConfSec>& Sections(void) const{return sec;}
        TConfSec& operator[](int i){return sec[i];}
        const TConfSec& at(int i) const{return sec.at(i);}
        // Display DOM
        void Print(void) const;
        // Get number of sector with the given name
        int FindSec(const string& sec_name) const;
        // Get assignment number of the given variable in the given sector
        int FindAssgn(int sec_num, const string& var_name) const;
        // Determine sector number and assignment number for the given
        // name of the sector and name of the variable
        bool Find(const string& sec_name, const string& var_name, int& sec_num,
                int& assgn_num) const;
        // Add sector to DOM
        void AddSec(const TConfSec& s){sec.push_back(s);}
        // Inser sec to DOM
        bool InsertSec(int index, const TConfSec& s);
        // Add assignment to the last sector
        void AddAssgn(vector<string>& a){sec[sec.size() - 1].AddAssgn(a);}
        // Add assignment to the given sector
        bool PushAssgnToSec(vector<string>& a, const string& sec_name,
                bool force=false);
        // Write sec with given index to the s object ref
        bool GetSec(int i, TConfSec& s) const;
        // Write sec with given name to the s object ref
        bool GetSec(const string &sec_name, TConfSec& s) const;
        // Write last/top DOM section to s object ref
        bool GetLastSec(TConfSec& s) const;
        // Drop sector with the given index
        bool DropSec(int index);
        // Drop last sector
        bool DropLastSec(void){return DropSec(sec.size() - 1);}

        void Clear(void){sec.clear();}
        TConfDOM(void){}
        ~TConfDOM(void){}
    };

//-----------------------------------------------------------------------------

// Configuration file handler
class TConf{
    private:
        string file_path;   // Full path to configuration file
        TFileInMem file;    // Conf-file in memory
        TConfDOM dom;       // Configuration DOM
        string error;       // Last error message

        // Assign indexes of opening and closing quotes
        bool FindQtnMarks(const string& line, int& start, int& stop) const;
        // Assign next quote from given line to 'word' and assign indexes
        // of opening and closing quotes
        bool GetQuote(const string& line, string& word, int& start,
                int& stop) const;
        // Pars given line of configuration file
        bool ParseNextLine(string& line, TConfSec& sec,
                vector<string>& assgn) const;
        // Search for and get sector name from the given line
        string NextSec(const string& line) const;
        // Search for and get next assignment from the given line
        vector<string> NextAssgn(string& line) const;
        // Form  assignment string
        string AssgnStr(const vector<string>& assgn) const;
        // Add new line to the conf file data string
        bool AddNewLine(int next_line_n, const string& text);
        // Add new assignment to the conf file data string
        bool AddNewAssgn(int next_line_n, const vector<string>& new_assgn){
            return AddNewLine(next_line_n, AssgnStr(new_assgn) + "\n");
            }
        // Replace given line of the Conf-file data string with given assgn
        bool ReplaceLine(int n, const vector<string>& new_assgn);

    public:
        // Display Conf-file data
        void PrintFile(void) const{
            cout << "Configuration file contents:" << endl << file.data;
            }
        // Display configuration DOM
        void PrintDOM(void) const{dom.Print();}
        // Get full path of the configuration file
        const string& FilePath(void){return file_path;}
        // Write full path of the configuration file
        void WriteFilePath(const string& path=""){file_path=path;}
        // Read Conf-file to memory
        bool ReadFile(const string &path="");
        // Parse Conf-file and form DOM
        bool Parse(void);
        // Get vector of all the sections of the conf data
        const vector<TConfSec>& AllSec(void) const{return dom.Sections();}
        // Get vector of names of all the sections of the conf data
        vector<string> AllSecNames(void) const;
        // Get vector of names of the variables of the sector
        vector<string> AllVarNamesOfSec(const string& sec_name="");
        // Read first value of the named variable from given sector into param
        bool ReadDOMParam(string& param, const TConfSec& sec,
                const string& var) const;
        // Read first value of the named variable from named sector into param
        bool ReadDOMParam(string& param, const string& sec,
                const string& var) const;
        // Read values of the given varibales of the given sector into vparam
        bool ReadVParams(vector<string>& vparam, const string& sec,
                const string& var) const;
        // Write value of the named variable in the named sector
        bool ChangeDOMParam(const string& val, const string& sec,
                const string& var, bool force=false);
        // Add sector to DOM of the conf file
        void AddDOMSec(const TConfSec& s){dom.AddSec(s);}
        // Insert sector to DOM of the conf file
        bool InsertDOMSec(int index, const TConfSec& sec){
            return dom.InsertSec(index, sec);
            }
        // Clear DOM of the conf file
        void CleanDOM(void){dom.Clear();}
        // Drop sector with the given index from the DOM of the conf
        bool DropDOMSec(int index){return dom.DropSec(index);}
        // Drop last sector from the DOM of the conf
        bool DropLastDOMSec(void){return dom.DropLastSec();}
        // Get number of DOM sections
        int DOMSize(void){return dom.Sections().size();}
        // Rewrite configuration file completly according to DOM
        bool RewriteAccordingToDOM(const string& message="");
        // Write new assignent back into Conf-file data string
        bool WriteParam(const string& sec, const string& var,
                bool write_to_file=true);
        // Check whether sector with the given name exists
        bool CheckSec(const string& sec_name) const{
            return (dom.FindSec(sec_name) >= 0);
            }
        // Get DOM section with given index
        bool GetDOMSec(int i, TConfSec& s) const{
            return dom.GetSec(i, s);
            }
        // Get DOM section with given name
        bool GetDOMSec(const string &sec_name, TConfSec& s) const{
            return dom.GetSec(sec_name, s);
            }
        // Get last DOM section
        bool GetLastDOMSec(TConfSec& s) const{return dom.GetLastSec(s);}

        TConf(void){}
        ~TConf(void){}
    };

//-----------------------------------------------------------------------------

#endif
