/*
    control++ : System configuration tool
    Copyright (C) 2017-2019 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef permissionsH
#define permissionsH

#include "permissions_basic.h"
#include "filesystem.h"

#define ACL_SUFFIX      "_acl"      // Suffix of the ACL perms sectors

#define TYPE_FPERM      0
#define TYPE_FACL       1

#define FPERM_PARAMS_N  4
#define FPERM_PATH      0
#define FPERM_OWNER     1
#define FPERM_GROUP     2
#define FPERM_MODE      3

#define FACL_PARAMS_N   5
#define FACL_PATH       0
#define FACL_USER       1
#define FACL_GROUP      2
#define FACL_MASK       3
#define FACL_OTHER      4

//-----------------------------------------------------------------------------

// Regular permission mode or ACL
class TFileMode;
// Permission of particular file
class TFilePerm;
// Permission of group of files
class TGroupOfFilesPerm;
// ACL of particular file
class TFileACL;
// ACL of group of files
class TGroupOfFilesACL;

//-----------------------------------------------------------------------------

// Get current permission of the file
TFilePerm GetPermission(const string& file_path, string& error);

// Get current ACL of the file
TFileACL GetACL(const string& file_path, string& error);

// Checking controversy between given modes
bool Controversy(TFileMode *m, TFileMode *m2);

//-----------------------------------------------------------------------------

// Description of sector type
class TSectorType{
    private:
        bool valid = false;
        bool file = false;
        bool dir = false;
        bool colorlist = false;
        bool whitelist = false;
        bool blacklist = false;
        bool acl = false;
        bool recursive = false;
    public:
        // Fields access
        bool Valid(void) const{return valid;}
        bool File(void) const{return file;}
        bool Dir(void) const{return dir;}
        bool List(void) const{return (colorlist || whitelist || blacklist);}
        bool WhiteList(void) const{return whitelist;}
        bool BlackList(void) const{return blacklist;}
        bool ACL(void) const{return acl;}
        bool Recursive(void) const{return recursive || whitelist || blacklist;}
        // Determine exact type of sector by it's name
        void FormType(const string &sector_name);
        // Sector for the dir permission mode
        bool DirSector(const string &sector_name, const string &suffix="");
        // Sector type is whitelist/blacklist or just list
        bool BlackListSector(const string &sector_name, const string &suffix="");
        bool WhiteListSector(const string &sector_name, const string &suffix="");
        bool ColorListSector(const string &sector_name, const string &suffix="");
        // Reset the data
        void Clear(void){
            valid = file = dir = colorlist = whitelist = blacklist
                = acl = recursive = false;
            }
    };

//-----------------------------------------------------------------------------

// Regular permission mode or ACL
class TFileMode{
    protected:
        string path;    // Full file path
        map<string, string> attrs;   // Attributes of a file
        const map<string, string> *p_attrs = NULL;
    public:
        // Get the type id
        virtual int Type(void) const=0;
        // Spawn child object and return the pointer
        TFileMode* Spawn(bool store_pointer_only=false);
        // Get/write file path
        const string& Path(void) const{return path;}
        bool WritePath(const string &val){
            if(!val.empty()){
                path = val;
                return true;
                }
            return false;
            }
        // Get main data fields
        virtual vector<string> GetData(void)=0;
        // Check if file is accessible
        bool Access(void){return (access(path.c_str(), F_OK) != -1);}
        // Assgn given data field with given value
        virtual bool Assgn(const string &var, const string &val)=0;
        // Revise the mode
        virtual bool Revise(const TSectorType& sector, string &error)=0;
        // Apply mode of the file
        virtual bool Apply(string& error, string* new_perm=NULL)=0;
        // Check mode compliance of the file
        virtual bool CheckCompliance(string &error, string &warning)=0;
        // Copy the mode without any checks
        virtual void CopyMode(const vector<string> &m)=0;
        virtual void CopyMode(const string &m)=0;
        // Check that every data field is not empty
        virtual bool Complete(void) const=0;
        // Check that mode can't have any effect
        virtual bool Neutral(void) const=0;

        // Print params of the mode
        virtual void Print(ostream &out) const=0;
        void Print(void) const{
            Print(cout);
            }
        // Clear all data fields
        virtual void Clear(void);

        TFileMode(void){};
        virtual ~TFileMode(void){};
    };

//-----------------------------------------------------------------------------

// Permission for a particular file
class TFilePerm : public TFileMode{
    protected:
        // Copy data of the permission object
        void CopyFilePerm(const TFilePerm &f);
        // Copy data stored by p_attrs to attrs vector
        void PullData(void);
        // Chown ang chgrp of the file
        bool ChangeOwnerAndGroup(string& error, TFilePerm* new_perm=NULL);
        // Chmod of the file
        bool ChangePermMode(string& error, TFilePerm* new_mode=NULL);
    public:
        // Get the type id
        int Type(void) const{return TYPE_FPERM;}
        // Fields access
        const string& Owner(void) const;
        const string& Group(void) const;
        const string& Mode(void) const;
        string OwnerMode(void) const;
        string GroupMode(void) const;
        string OtherMode(void) const;
        // Get main data fields
        vector<string> GetData(void);
        // Modifiers
        bool WriteOwner(const string &val){
            attrs["owner"] = val;
            return true;
            }
        bool WriteGroup(const string &val){
            attrs["group"] = val;
            return true;
            }
        bool WriteMode(const string &val){
            if(CheckModeString(val)){
                attrs["mode"] = val;
                return true;
                }
            return false;
            }
        // Copy the mode without any checks
        void CopyMode(const vector<string> &m);
        void CopyMode(const string &m);
        bool WriteEverything(const string &p, const string &o, const string &g,
                const string &m);
        // Assgn given data field with given value
        bool Assgn(const string &var, const string &val);
        // Revise the mode
        bool Revise(const TSectorType& sector, string &error);
        // Apply mode of the file
        bool Apply(string& error, string* new_perm=NULL);
        // Check according bits of the mode field of the particular file stat
        bool CheckCompliance(string &error, string &warning);
        // Check controversy with the given mode
        bool Controversy(const TFilePerm& other_perm);
        // Check that every data field is not empty
        bool Complete(void) const;
        // Check that mode can't have any effect
        bool Neutral(void) const;
        // Print current permission mode
        void Print(ostream &out) const;
        // Clear all data fields
        void Clear(void){
            TFileMode::Clear();
            }

        TFilePerm(void){
             attrs["owner"];
             attrs["group"];
             attrs["mode"];
             p_attrs = &attrs;
             }
        TFilePerm(const string &p, const string &o, const string &g,
                const string &m);
        TFilePerm(const map<string, string> *attributes);
        // Copy constructor and assgn operator
        TFilePerm(const TFilePerm &f);
        TFilePerm &operator=(const TFilePerm &f);

        ~TFilePerm(void){}
    };

//-----------------------------------------------------------------------------

// ACL of a file
class TFileACL : public TFileMode{
    protected:
        map<string, string> user_q;     // Qualified user
        map<string, string> group_q;    // Qualified group

        map<string, string> empty_map;

        // Copy data of the permission object
        void CopyFileACL(const TFileACL &f);
        // Modify given data field
        bool WriteAttr(string &str_attr, map<string, string> &map_attr,
                const string& attr_val, const string &q="",
                bool mode_str=true);
    public:
        // Get the type id
        int Type(void) const{return TYPE_FACL;}
        // Fields access
        const string& User(void) const{return attrs.at("user");}
        const map<string, string>& QUsers(void) const{return user_q;}
        const string UserQ(const string& q) const{
            if(user_q.find(q) != user_q.end())
                return user_q.at(q);
            return "";
            }
        const string& Group(void) const{return attrs.at("group");}
        const map<string, string>& QGroups(void) const{return group_q;}
        const string GroupQ(const string& q) const{
            if(group_q.find(q) != group_q.end())
                return group_q.at(q);
            return "";
            }
        const string& Mask(void) const{return attrs.at("mask");}
        const string& Other(void) const{return attrs.at("other");}
        // Get main data fields
        vector<string> GetData(void);
        // Modifiers
        bool WriteUser(const string &val, const string &q=""){
            return WriteAttr(attrs["user"], user_q, val, q);
            }
        bool WriteGroup(const string &val, const string &q=""){
            return WriteAttr(attrs["group"], group_q, val, q);
            }
        bool WriteMask(const string &val){
            return WriteAttr(attrs["mask"], empty_map, val);
            }
        bool WriteOther(const string &val){
            return WriteAttr(attrs["other"], empty_map, val);
            }
        bool WriteQUsers(const map<string, string> &u_q);
        bool WriteQGroups(const map<string, string> &g_q);
        bool WriteNecessary(const string &p, const string &u, const string &g,
                const string &m, const string &o);
        bool WriteEverything(const string &p, const string &u,
                const map<string, string> &u_q, const string &g,
                const map<string, string> &g_q, const string &m,
                const string &o);
        // Write ACL param identified by given tag
        bool WriteByTag(acl_tag_t tag, const string &val, const string &q="");
        // Copy the mode without any checks
        void CopyMode(const vector<string> &m);
        void CopyMode(const string &m);
        // Assgn given data field with given value
        bool Assgn(const string &var, const string &val);
        // Revise the mode
        bool Revise(const TSectorType& sector, string &error);
        // Apply ACL of the file
        bool Apply(string& error, string* new_perm=NULL);
        // Check ACL compliance of the file
        bool CheckCompliance(string &error, string &warning);
        // Check controversy with the given mode
        bool Controversy(const TFileACL& other_acl);
        // Check that every data field is not empty
        bool Complete(void) const;
        // Check that mode can't have any effect
        bool Neutral(void) const;
        // Print current permission mode
        void Print(ostream &out) const;
        // Clear all data fields
        void Clear(void);

        TFileACL(void){
            attrs["user"];
            attrs["group"];
            attrs["mask"];
            attrs["other"];
            p_attrs = &attrs;
            }
        TFileACL(const string &p, const string &u, const string &g,
                const string &m, const string &o);
        TFileACL(const string &p, const string &u,
                const map<string, string> &u_q, const string &g,
                const map<string, string> &g_q, const string &m,
                const string &o);
        TFileACL(const map<string, string> *attributes);
        // Copy constructor and assgn operator
        TFileACL(const TFileACL &f);
        TFileACL &operator=(const TFileACL &f);

        ~TFileACL(void){}
        };

//-----------------------------------------------------------------------------

// Additional params for a group of files
// (dir or list of file paths stored in txt file)
class TGroupOfFilesMode{
    private:
        string nested_dirs_mode;        // Mode for the nested dirs
        string base_dir;                // Every other path is relative
                                        //  to base dir
        vector<string> excluded_paths;  // File that will be excluded
                                        //  from mode set
        // Modify given data field
        bool WriteAttr(string& attr, const string& attr_val,
                bool mode_attr=false);
    public:
        // Fields access
        const string& DirsMode(void) const{return nested_dirs_mode;}
        const string& BaseDir(void) const{return base_dir;}
        const vector<string>& ExcludedPaths(void) const{
            return excluded_paths;
            }
        vector<string> BaseDirExcludedPaths(void) const{
            return RebaseFilePaths(excluded_paths, base_dir);
            }
        // Modifiers
        bool WriteDirsMode(const string &val){
            return WriteAttr(nested_dirs_mode, val, true);
            }
        bool WriteBaseDir(const string &val){
            return WriteAttr(base_dir, FormatFilePath(val), false);
            }
        bool WriteExcludedPaths(vector<string> epaths);
        bool AddExcludedPath(const string& epath);
        bool WriteEverything(const string &p, const string &o, const string &g,
                const string &m, const string &md);
        // Assgn given data field with given value
        bool Assgn(const string &var, const vector<string> &vals);
        bool Assgn(const string &var, const string &val){
            return Assgn(var, vector<string>(1, val));
            }
        // Revise the mode
        bool Revise(const TSectorType& sector, string &error);
        // Check controversy with the given mode
        bool Controversy(const TFileACL& other_acl);
        // Check that every data field is not empty
        bool Complete(void) const;
        // Check that mode can't have any effect
        bool Neutral(void) const;
        // Clear all data fields
        void Clear(void);
    };

//-----------------------------------------------------------------------------

#endif
