/*
    control++ : System configuration tool
    Copyright (C) 2017-2019 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef printH
#define printH

#include "libcontrol++.h"
#include <vector>

#define INDENT  "\t"        // Using tabs for indentation
#define TERM_MAX_WIDTH  256 // Limit of value returned by the GetTerminalWidth

//-----------------------------------------------------------------------------

// Get terminal width
int GetTerminalWidth(void);

// Printer of the complex messages
class TPrinter{
    private:
        bool silent = false;            // No output
        bool silent0 = false;           // Prev value of 'silent'
        bool silent_on_begin = false;   // Value of 'silent flag at the moment of operation printing
        bool numb = false;              // Don't do a thing
        bool plain = false;             // Plain output
        bool instant_feedback = false;  // Print additional info right away for
                                        //  thecurrent operation
        char tie_symbol = '.';  // Symbol used to tie left and right parts
                                //  of the status line
        vector<string> indent;  // Indentation stack
        string operation;       // Left part of the status line
        string operation0;      // Prev operation
        vector<string> extra;   // Extra info displayed
                                //  under the main status line

        // Print according to current mode
        void Print(const string& str) const;
        // Print line of extra info
        void PrintExtraLine(const string& ex) const;
        // Print indentation units
        void PrintIndent(void);
        // Print given string at the right part of the terminal and
        //  tie it with the left part by the row of 'tie' symbols
        void CompleteTheLine(const string& status);
        // Form special message
        string SpecMessage(const string& preface, const string& message);
        // Clearing data fields
        void Clear(void);
    public:
        // Get current left part of the status line
        string Operation(void){return operation;}
        // Get vector of extra lines
        vector<string> Extra(void){return extra;}
        // Modifiers
        void Silent(bool flag){
            silent0 = silent;
            silent = flag;
            }
        void ReSilent(void){
            Silent(silent0);
            }
        void Numb(bool flag){numb = flag;}
        void Plain(bool flag){plain = flag;}
        void TieSymbol(char s){tie_symbol = s;}
        // Access to the mode-flags
        bool IsSilent(void){return silent;}
        bool IsPlain(void){return plain;}
        // Start a new status line
        void LineBegin(const string& message);
        // Complete current status line
        void LineEnd(const string& message="");
        // Add extra line, displayed under the main status line
        void LineAddExtra(const string& message);
        // Add extra error line
        void LineAddErr(const string& message=""){
            LineAddExtra(SpecMessage("ERROR", message));
            }
        // Add extra warning line
        void LineAddWarn(const string& message=""){
            LineAddExtra(SpecMessage("WARNING", message));
            }
        // Clear vector of extra messages and put the new one in
        void LineRewriteExtra(const string& message);
        // Break current status line
        void LineBreak(void);
        // Begin the new line with the same message on the left side
        void LineRestart(void);
        // Push / pop indentation symbol / unit, or drop them altogether
        void Indent(void){indent.push_back(INDENT);}
        void UnIndent(void){indent.pop_back();}
        void DropIndent(void){indent.clear();}
        // Display additional info right away,
        //  not after printing the completion status
        void StartInstantFeedback(const string& trivial_message = "[IN PROGRESS]");
        // Print error message
        void Err(const string& message=""){
            LineEnd(SpecMessage("ERROR", message));
            }
        // Print warning message
        void Warn(const string& message=""){
            LineEnd(SpecMessage("WARNING", message));
            }
        // Print message of seccessfull result
        void Success(const string& message=""){
            LineEnd(SpecMessage("DONE", message));
            }
        // Print divider on entire terminal line
        void Div(char s = '=') const;

    TPrinter(){
        cout << unitbuf; // Automaticly flushing the output
        }
    ~TPrinter(){}
    };

//-----------------------------------------------------------------------------

#endif
